#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include <iostream.h>


#include "elf.if"	// <-- 
#include "ord.if"
#include "aof.if"

//////////////////////////////////////////////////////////////////////////
//									//
//////////////////////////////////////////////////////////////////////////

Chunk::Chunk (void)
  {
    id[0] = 0;
    fileOffset = 0;
    size = 0;
  }

Chunk::Chunk (char *i, int o, int s)
  {
    strcpy (id,i);
    fileOffset = o;
    size = s;
  }

Chunk::~Chunk (void)
  {
  }

void Chunk::print (ostream &strm)
  {
    strm << "Chunk id: " << id << '\n';
    strm << "Offset:   " << asHex32X (fileOffset) << '\n';
    strm << "Size:     " << asHex32X (size) << '\n';
  }

void Chunk::parse (Reader *in)
  {
  }

Chunk *Chunk::virtualConstructor (Reader *in)
  {
    Chunk *ret;
    char tmp[9];
    int ofs;
    int siz;
    
    for (int i = 0; i < 8; ++i)
      {
	tmp[i] = in->get8 ();
      }
    tmp[8] = '\0';
    ofs = in->get32 ();
    siz = in->get32 ();
    if (strcmp (tmp, "OBJ_HEAD") == 0)
      {
	ret = new AOFHeaderChunk (tmp,ofs,siz);
      }
    else if (strcmp (tmp, "OBJ_AREA") == 0)
      {
	ret = new AreaChunk (tmp,ofs,siz);
      }
    else if (strcmp (tmp, "OBJ_SYMT") == 0)
      {
	ret = new SymbolChunk (tmp,ofs,siz);
      }
    else if (strcmp (tmp, "OBJ_IDFN") == 0)
      {
	ret = new IDChunk (tmp,ofs,siz);
      }
    else if (strcmp (tmp, "OBJ_STRT") == 0)
      {
	ret = new StringChunk (tmp,ofs,siz);
      }
    else
      {
	ret = new Chunk (tmp,ofs,siz);
      }

    ret->parse (in);

    return (ret);
  }
    
bool Chunk::isAOFHeaderChunk (void)
  {
    return (false);
  }


bool Chunk::isStringChunk (void)
  {
    return (false);
  }

bool Chunk::isSymbolChunk (void)
  {
    return (false);
  }

    
//////////////////////////////////////////////////////////////////////////
//									//
//////////////////////////////////////////////////////////////////////////

AreaHeader::AreaHeader (void)
  {
  }


AreaHeader::~AreaHeader (void)
  {
  }


void AreaHeader::parse (Reader *in)
  {
    nameOffset = in->get32 ();
    attributes = in->get32 ();
    areaSize   = in->get32 ();
    noOfRelocs = in->get32 ();
    baseAddress = in->get32 ();
  }

void AreaHeader::print (ostream &strm)
  {
    strm << "Name offset  " << nameOffset <<" (" <<  g_aof->stringChunk->getString (nameOffset) << ")\n";
    strm << "Attributes   " << asHex32X (attributes) << '\n';
    strm << "Area size    " << areaSize << '\n';
    strm << "Num Relocs   " << noOfRelocs << '\n';
    strm << "Base Address " << baseAddress << '\n';
  }
    
//////////////////////////////////////////////////////////////////////////
//									//
//////////////////////////////////////////////////////////////////////////

AOFHeaderChunk::AOFHeaderChunk (char *i, int o, int s)
  {
    strcpy (id,i);
    fileOffset = o;
    size = s;
  }


void AOFHeaderChunk::parse (Reader *in)
  {
    int save_pos = in->getPos ();	// Save file position

    in->setPos (fileOffset);
    
    objFileType = in->get32 ();
    versionId = in->get32 ();
    noOfAreas = in->get32 ();
    noOfSymbols = in->get32 ();
    entryAreaIndex = in->get32 ();
    entryOffset = in->get32 ();
    
    headers = new (AreaHeader (* [noOfAreas]));
    
    for (int i = 0; i < noOfAreas; ++i)
      {
	headers[i] = new AreaHeader ();
	headers[i]->parse (in);
      }
    
    in->setPos (save_pos);		// Restore file position
  }


AOFHeaderChunk::~AOFHeaderChunk (void)
  {
  }


void AOFHeaderChunk::print (ostream &strm)
  {
    Chunk::print (strm);
    strm << "File type:" << asHex32X (objFileType) << '\n';
    strm << "Version:  " << versionId << '\n';
    strm << "Areas:    " << noOfAreas << '\n';
    strm << "Symbols:  " << noOfSymbols << '\n';

    strm << "EntryIdx  " << entryAreaIndex << '\n';
    strm << "Offset    " << entryOffset << '\n';
    
    for (int i = 0; i < noOfAreas; ++i)
      {
	strm << "Area " << i+1 << '\n';
	headers[i]->print (strm);
      }

  }

bool AOFHeaderChunk::isAOFHeaderChunk (void)
  {
    return (true);
  }


int AOFHeaderChunk::getNoOfAreas (void)
  {
    return (noOfAreas);
  }

AreaHeader *AOFHeaderChunk::getArea (int area_no)
  {
    return (headers[area_no]);
  }

int AOFHeaderChunk::getAreaSize (int area_no)
  {
    return (headers[area_no]->areaSize);
  }


int AOFHeaderChunk::getNoOfRelocs (int area_no)
  {
    return (headers[area_no]->noOfRelocs);
  }


//////////////////////////////////////////////////////////////////////////
//									//
//////////////////////////////////////////////////////////////////////////

AreaContent::AreaContent (size_t s)
  {
    size = s;
    bytes = new char [size];
  }

void AreaContent::parse (Reader *in)
  {
    for (int i = 0; i < size; ++i)
      {
	bytes[i] = in->get8 ();
      }
  }

void AreaContent::dumpArea (int size, ostream &strm)
  {
    strm << asHex32 (0) << " ";
    
    for (int i = 0; i < size; ++i)
      {
	if (i && (i % 16 == 0))
	  strm << '\n' << asHex32 (i) << " ";
	strm << ' ' << asHex8 (bytes[i+pos] & 0xFF);
      }
    strm << '\n';
  }

void AreaContent::dumpRelocs (int relocs, ostream &strm)
  {
    for (int i = 0; i < relocs; ++i)
      {
	int j = helper.makeWord (bytes[i*8+pos],
				 bytes[i*8+pos+1],
				 bytes[i*8+pos+2],
				 bytes[i*8+pos+3]);
	strm << "offset: ";
	strm << asHex32X (j);
	strm << "  ";
	strm << "relocation: ";
	j = helper.makeWord (bytes[i*8+pos+4],
			     bytes[i*8+pos+5],
			     bytes[i*8+pos+6],
			     bytes[i*8+pos+7]);
	strm << asHex32X (j);
	switch ( (j >> 25) & 0x3)
	  {
	    case 0:
	      strm << " byte";
	      break;
	    case 1:
	      strm << " halfword";
	      break;
	    case 2:
	      strm << " word";
	      break;
	    case 3:
	      strm << " instruction";
	      break;
	  }
	if ((j >> 27) & 0x01)
	  {
	    strm << ' ' << g_aof->stringChunk->getString (
	      g_aof->symbolChunk->symbols [j]->name);
	  }
	else
	  {
	    strm << " base of area " << (j & 0x0FFFFFF);
	  }
	if ((j >> 26) & 0x01)
	  {
	    strm << " R";
	  }
	if ((j >> 28) & 0x01)
	  {
	    strm << " B";
	  }
		
	strm << '\n';
      }
  }


void AreaContent::print (ostream &strm)
  {
    int areas;
    pos = 0;

    areas = g_aof->headerChunk->getNoOfAreas ();

    for (int i = 0; i < areas; ++i)
      {
	int size;
	AreaHeader *a;

	a = g_aof->headerChunk->getArea (i);
	size = g_aof->headerChunk->getAreaSize (i);

	strm << "Area "<< i << " (";
	strm << g_aof->stringChunk->getString (a->nameOffset);
	strm << ")\n";

	 strm << "size: " << size << '\n';
	 
	if (a->attributes & 0x1000)
	  {
	    strm << "-no init-\n";
	  }
	else
	  {
	    dumpArea (size, strm);

	    pos += size;
	  }
	
	strm << "Relocs\n";
	
	int relocs = g_aof->headerChunk->getNoOfRelocs (i);
	if (relocs == 0)
	  {
	    strm << "-none-\n";
	  }
	else
	  {
	    dumpRelocs (relocs, strm);
	    pos += 8*relocs;
	  }
	strm << '\n';
      }
  }


AreaChunk::AreaChunk (char *i, int o, int s)
  {
    strcpy (id,i);
    fileOffset = o;
    size = s;
  }


void AreaChunk::parse (Reader *in)
  {
    int save_pos = in->getPos ();	// Save file position

    in->setPos (fileOffset);

    areas = new AreaContent (size);
    
    areas->parse (in);
    
    in->setPos (save_pos);		// Restore file position
  }


AreaChunk::~AreaChunk (void)
  {
  }


void AreaChunk::print (ostream &strm)
  {
    Chunk::print (strm);
    strm << "\n";
    areas->print (strm);
  }


//////////////////////////////////////////////////////////////////////////
//									//
//////////////////////////////////////////////////////////////////////////

SymbolContent::SymbolContent (void)
  {
  }

void SymbolContent::parse (Reader *in)
  {
    name = in->get32 ();
    attributes = in->get32 ();
    value = in->get32 ();
    areaName = in->get32 ();
  }

void SymbolContent::print (ostream &strm)
  {
    strm << "Symbol: " << g_aof->stringChunk->getString (name) << " [" << name << "]\n";
    strm << "Attrib: " << asHex32X (attributes) << '\n';
    strm << "Value:  " << value << '\n';
    strm << "Area:   " << areaName << '\n';
  };
  
SymbolChunk::SymbolChunk (char *i, int o, int s)
  {
    strcpy (id,i);
    fileOffset = o;
    size = s;
  }


void SymbolChunk::parse (Reader *in)
  {
    int save_pos = in->getPos ();	// Save file position

    in->setPos (fileOffset);

    symbols = new (SymbolContent (* [size/16 + 1]));
    
    for (int i = 0; i < size/16; ++i)
      {
	symbols[i] = new SymbolContent ();
	symbols[i]->parse (in);
      }
    
    in->setPos (save_pos);		// Restore file position
  }


SymbolChunk::~SymbolChunk (void)
  {
  }


void SymbolChunk::print (ostream &strm)
  {
    Chunk::print (strm);
    for (int i = 0; i < size/16; ++i)
      {
	strm << "Symbol " << i+1 << '\n';
	symbols[i]->print (strm);
      }

  }


bool SymbolChunk::isSymbolChunk (void)
  {
    return (true);
  }


//////////////////////////////////////////////////////////////////////////
//									//
//////////////////////////////////////////////////////////////////////////

IDChunk::IDChunk (char *i, int o, int s)
  {
    strcpy (id,i);
    fileOffset = o;
    size = s;
    idString = new char [s];
  }


void IDChunk::parse (Reader *in)
  {
    int save_pos = in->getPos ();	// Save file position

    in->setPos (fileOffset);

    for (int i = 0; i < size; ++i)
      {
	idString[i] = in->get8 ();
//	cout << idString[i];
      }
    
    in->setPos (save_pos);		// Restore file position
  }


IDChunk::~IDChunk (void)
  {
    delete [] idString;
  }


void IDChunk::print (ostream &strm)
  {
    Chunk::print (strm);

    strm << "ID: " << idString << '\n';
  }


//////////////////////////////////////////////////////////////////////////
//									//
//////////////////////////////////////////////////////////////////////////

  
StringChunk::StringChunk (char *i, int o, int s)
  {
    strcpy (id,i);
    fileOffset = o;
    size = s;
  }


void StringChunk::parse (Reader *in)
  {
    int save_pos = in->getPos ();	// Save file position

    in->setPos (fileOffset);

    length = in->get32 ();

    stringTable = new char [length];
    
    for (int i = 0; i < length - 4 ; ++i)
      {
	stringTable[i] = in->get8 ();
      }
    
    in->setPos (save_pos);		// Restore file position
  }


StringChunk::~StringChunk (void)
  {
  }


void StringChunk::print (ostream &strm)
  {
    Chunk::print (strm);

    int currpos = 0;
    
    for (int j = 0; currpos < length - 4; ++j)
      {
	strm << "String " << j+1 << "[" << currpos + 4 <<  "]: ";
	
	for (int i = currpos; i < length; ++i)
	  {
	    if (stringTable[i] == '\0')
	      {
		break;
	      }
	    strm << stringTable[i];
	  }
	currpos = i + 1;
	strm << '\n';
      }
  }

char *StringChunk::getString (int offset)
  {
    if (offset > length || offset < 4)
      return ("-no string-");
    
    return (&stringTable[offset - 4]);
  }


bool StringChunk::isStringChunk (void)
  {
    return (true);
  }


//////////////////////////////////////////////////////////////////////////
//									//
//////////////////////////////////////////////////////////////////////////

AOFFile::AOFFile (Reader *in)
  {
    unsigned long l;
    
    rd = in;

    l = in->get32raw ();

    if (l == 0xC3CBC6C5)
      bigEndian = true;
    else
      bigEndian = false;

    in->bigEndian = bigEndian;
    
    maxChunks = in->get32 ();
    numChunks = in->get32 ();

    int no = numChunks;

    chunkPool = new (Chunk (* [numChunks]));
    
    for (int i = 0; i < numChunks; ++i)
      {
	chunkPool[i] = Chunk::virtualConstructor (in);
	
	if (chunkPool[i]->isAOFHeaderChunk ())
	  {
	    headerChunk = (AOFHeaderChunk *) (chunkPool[i]);
	  }
	if (chunkPool[i]->isStringChunk ())
	  {
//	    stringChunk = dynamic_cast <StringChunk *> (chunkPool[i]);
	    stringChunk = (StringChunk *) (chunkPool[i]);
	  }
	if (chunkPool[i]->isSymbolChunk ())
	  {
	    symbolChunk = (SymbolChunk *) (chunkPool[i]);
	  }
      }
  }



void AOFFile::print (ostream &strm)
  {
    if (bigEndian)
      {
	helper.setBigEndian ();
	strm << "big-endian\n";
      }
    else
      {
	strm << "low-endian\n";
      }

    strm << "Max chunks: " << maxChunks << '\n';
    strm << "Num chunks: " << numChunks << '\n';

    if (chunkPool)
      {
	for (int i = 0; i < numChunks; ++i)
	  {
	    strm << '\n';
	    chunkPool[i]->print (strm);
	  }
      }
  }


