#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include <iostream.h>

#include <io.h>

#include "aof.if"
#include "elf.if"
#include "ord.if"

extern char *disassemble (unsigned w, unsigned long base);

void error (const char *fmt, ...);

AOFFile *g_aof;
ElfFile *g_elf;

Helper helper;

char *stringer;
char *shtringer;

SHSection *sh_sect;
SymbolTable *sym_tab;


//////////////////////////////////////////////////////////////////////////
//									//
//////////////////////////////////////////////////////////////////////////

char *asHex (char *fmt, int i)
  {
    static char b[32];

    sprintf (b, fmt, i);

    return (b);
  }

char *asHex8X (int i)
  {
    return (asHex ("0x%02X",i & 0xFF));
  }

char *asHex16X (int i)
  {
    return (asHex ("0x%04X",i & 0xFFFF));
  }

char *asHex32X (int i)
  {
    return (asHex ("0x%08X",i));
  }

char *asHex8 (int i)
  {
    return (asHex ("%02X",i & 0xFF));
  }

char *asHex16 (int i)
  {
    return (asHex ("%04X",i & 0xFFFF));
  }

char *asHex32 (int i)
  {
    return (asHex ("%08X",i));
  }

char asChar (int i)
  {
    if (i >= ' ' && i < 0x7f)
      {
	return (i);
      }
    else
      {
	return ('.');
      }
  }


//////////////////////////////////////////////////////////////////////////
//									//
//////////////////////////////////////////////////////////////////////////

FileBuff::FileBuff (char *file_name)
  {
    if ((theFile=fopen(file_name,"rb")) == 0)
      {
	::error ("Cannot Open %s",file_name);
	exit(1);
      }
    fileSize = _filelength (_fileno (theFile));

    buff = new char [fileSize + 3];

    fread (buff, 1, fileSize, theFile);
    pos = 0;
  }


FileBuff::FileBuff (char *array, int size)
  {
    buff = new char [size];

    for (int i = 0; i < size; ++i)
      {
/*
	if (i%16 == 0 && i)
	  cout << '\n';
	cout << asHex8 (array[i]) << " ";
*/
	buff[i] = array[i];
      }
    
    pos = 0;
    fileSize = size;
  }

int FileBuff::get (void)
  {
    if (pos < 0 || pos >= fileSize)
      {
	return (EOF);
      }

    int i = buff[pos] & 0xFF;
    
//    cout << "|" << hex <<  i << "|\n";

    return ( i );
  }

void FileBuff::advance (void)
  {
    pos = pos + 1;
  }

int FileBuff::getPos (void)
  {
    return (pos);
  }

void FileBuff::setPos (int i)
  {
    pos = i;
  }

char *FileBuff::getStringAt (int index)
  {
    return (&buff[index]);
  }

//////////////////////////////////////////////////////////////////////////
//									//
//////////////////////////////////////////////////////////////////////////

Reader::Reader (FileBuff *fb)
  {
    input = fb;
    bigEndian = false;
  }

int Reader::get8 (void)
  {
    int i = input->get ();

    input->advance ();
    
    return (i);
  }

int Reader::get16 (void)
  {
    int i;

    i = get8 ();
    if (bigEndian)
      i = i * 256 + get8 ();
    else
      i = i + get8 () * 256;

    return (i);
  }

int Reader::get16raw (void)
  {
    int i;

    i = get8 ();
    i = i * 256 + get8 ();

    return (i);
  }

int Reader::get32 (void)
  {
    int i;

    i = get16 ();
    if (bigEndian)
      i = i * 256 * 256 + get16 ();
    else
      i = i + get16 () * 256 * 256;

    return (i);
  }

int Reader::get32raw (void)
  {
    int i;

    i = get16raw ();
    i = i * 256 * 256 + get16raw ();

    return (i);
  }

int Reader::getPos (void)
  {
    return (input->getPos ());
  }

void Reader::setPos (int i)
  {
    input->setPos (i);
  }


//////////////////////////////////////////////////////////////////////////
//									//
//////////////////////////////////////////////////////////////////////////

Dissectionist::Dissectionist (void)
  {
  }

void Dissectionist::print (ostream &strm)
  {
    strm << "Dissectionist::print\n";
  }

//////////////////////////////////////////////////////////////////////////
//									//
//////////////////////////////////////////////////////////////////////////

SymbolTable::SymbolTable (char *arr, int size, int tabs)
  {
    FileBuff *sfb = new FileBuff (arr, size);

    Reader *srd = new Reader (sfb);

    nums = tabs;

    tables = new (Elf32_Sym (* [nums]));

    srd->bigEndian = false;

    for (int i = 0; i < nums; ++i)
      {
	tables[i] = new Elf32_Sym;
	tables[i]->st_name = srd->get32 ();
	tables[i]->st_value = srd->get32 ();
	tables[i]->st_size = srd->get32 ();
	tables[i]->st_info = srd->get8 ();
	tables[i]->st_other = srd->get8 ();
	tables[i]->st_shndx = srd->get16 ();
      }
  }

char *SymbolTable::getName (int index)
  {
    Elf32_Sym *sym = tables[index];

    return (&stringer[sym->st_name]);
  }
char *SymbolTable::getSection (int index)
  {
    char *p;
    Elf32_Sym *sym = tables[index];
    
    if (sym->st_shndx == SHN_UNDEF)
      p = " -undefined- ";
    else if (sym->st_shndx < SHN_LORESERVE && sh_sect)
      p = sh_sect->getShAt (sym->st_shndx);
    else
      p = "-reserved section-";

    return (p);
  }


void SymbolTable::print (ostream &strm)
  {
    strm << "  " << nums << " symbols:\n";
    for (int i = 0; i < nums; ++i)
      {
	Elf32_Sym *sym = tables[i];
	strm << "    Symbol: " << asHex8(i) << "\n";
	strm << "     st_name  " << asHex32 (tables[i]->st_name);
	strm << "      " << &stringer[sym->st_name] << '\n';
	
	strm << "     st_value " << asHex32 (sym->st_value) << '\n';
	strm << "     st_size  " << asHex32 (tables[i]->st_size) << '\n';
	strm << "     st_info  " << asHex8 (sym->st_info) << '\n';
	switch (ELF32_ST_BIND (sym->st_info))
	  {
	    case STB_LOCAL:
	      strm << "        local\n";
	      break;
	    case STB_GLOBAL:
	      strm << "        global\n";
	      break;
	    case STB_WEAK:
	      strm << "        weak\n";
	      break;
	    default:
	      ;
	  }
	
	strm << "     st_other " << asHex8 (tables[i]->st_other) << '\n';
	strm << "     st_shndx " << asHex16 (sym->st_shndx);
	if (sym->st_shndx == SHN_UNDEF)
	  strm << " -undefined- ";
	else if (sym->st_shndx < SHN_LORESERVE && sh_sect)
	  strm << " " << sh_sect->getShAt (sym->st_shndx);
	else
	  strm << "  reserved section";
	strm << "\n\n";
      }
  }


//////////////////////////////////////////////////////////////////////////
//									//
//////////////////////////////////////////////////////////////////////////

TextSection::TextSection (char *arr, int size)
  {
    FileBuff *sfb = new FileBuff (arr, size);

    Reader *srd = new Reader (sfb);

    nums = size / 4;

    codes = new unsigned[nums];

    srd->bigEndian = false;
    
    for (int i = 0; i < nums; ++i)
      {
	codes[i] = srd->get32 ();
      }
  }

void TextSection::print (ostream &strm)
  {
    int base = 0;
    strm << "  " << nums << " instructions:\n";
    for (int i = 0; i < nums; ++i)
      {
	strm << "    " << disassemble (codes[i], base);
	base += 4;
	strm << '\n';
      }
  }


//////////////////////////////////////////////////////////////////////////
//									//
//////////////////////////////////////////////////////////////////////////

SHSection::SHSection (char *arr, int size)
  {
    FileBuff *sfb = new FileBuff (arr, size);

    Reader *srd = new Reader (sfb);

    nums = 0;

    strings = new char [size];

    srd->bigEndian = false;

    char *p = arr;
    
    for (int i = 0; i < size; ++i)
      {
	strings[i] = *p;
	if (*p == 0)
	  ++nums;
	++p;
      }
    indxs = new int [nums];

    int k = 1;
    p = arr;

    indxs[0] = 0;
    
    for (int j = 0; j < size; ++j)
      {
	if (p[j] == '\0')
	  {
	    indxs[k] = j+1;
	    ++k;
	    if (k >= nums)
	      break;
	  }
      }
    
  }

char *SHSection::getShAt (int indx)
  {
    return (&strings[indxs[indx]]);
  }

void SHSection::print (ostream &strm)
  {
    strm << "  " << nums << " shstrings:\n";
    for (int i = 0; i < nums; ++i)
      {
	strm << "    [" << i << "] " << &strings[indxs[i]]; 
	strm << " [" << indxs[i] << "]";
	strm << '\n';
      }
  }


//////////////////////////////////////////////////////////////////////////
//									//
//////////////////////////////////////////////////////////////////////////

RelSection::RelSection (char *arr, int size, int ent_size)
  {
    FileBuff *sfb = new FileBuff (arr, size);

    Reader *srd = new Reader (sfb);

    nums = 0;

    if (ent_size != 8)
      return;

    nums = size/ent_size;
    
    relocs = new (Elf32_Rel (* [nums]));

    srd->bigEndian = false;

    for (int i = 0; i < nums; ++i)
      {
	relocs[i] = new Elf32_Rel;
	relocs[i]->r_offset = srd->get32 ();
	relocs[i]->r_info = srd->get32 ();
      }
  }

void RelSection::print (ostream &strm)
  {
    strm << "  " << nums << " relocations:\n";
    for (int i = 0; i < nums; ++i)
      {
	Elf32_Rel *rel = relocs[i];
	int sym_index;

	sym_index = ELF32_R_SYM (rel->r_info);
	
	strm << "    Relocation:\n";
	strm << "     r_offset " << asHex32 (rel->r_offset) << '\n';
	
	strm << "     r_info   " << asHex32 (rel->r_info) << '\n';
	strm << "        symbol " << sym_index;
	if (sym_tab)
	  {
	    strm << " ";
	    strm << sym_tab->getName (sym_index);
	    strm << " ";
	    strm << sym_tab->getSection (sym_index);
	  }
	if (g_elf)
	  {
	    strm << " | ";
	    if (sym_index < g_elf->numHdrs)
	      strm << g_elf->secHdrs[sym_index]->getName ();
	  }

	strm << "  [";
	switch (ELF32_R_TYPE (rel->r_info))
	  {
	    case R_ARM_NONE:
	      strm << "R_ARM_NONE";
	      break;
	    case R_ARM_PC24:
	      strm << "R_ARM_PC24";
	      break;
	    case R_ARM_ABS32:
	      strm << "R_ARM_ABS32";
	      break;
	    case R_ARM_REL32:
	      strm << "R_ARM_REL32";
	      break;
	    case R_ARM_PC13:
	      strm << "R_ARM_PC13";
	      break;
	    default:
	      strm << ELF32_R_TYPE (rel->r_info) << " -unknown-";
	  }
	      
	strm << "]\n";
      }
    strm << '\n';
  }


//////////////////////////////////////////////////////////////////////////
//									//
//////////////////////////////////////////////////////////////////////////

SectionHeader::SectionHeader (void)
  {
    dis = 0;
    data = 0;
  }

bool SectionHeader::parse (Reader *rd)
  {
    filePos = rd->getPos ();
    
    d.sh_name		= rd->get32 ();
    d.sh_type		= rd->get32 ();
    d.sh_flags		= rd->get32 ();
    d.sh_addr		= rd->get32 ();
    d.sh_offset		= rd->get32 ();
    d.sh_size		= rd->get32 ();
    d.sh_link		= rd->get32 ();
    d.sh_info		= rd->get32 ();
    d.sh_addralign	= rd->get32 ();
    d.sh_entsize	= rd->get32 ();

    if (d.sh_type != SHT_NOBITS)
      {
	int size = d.sh_size;
	
	data = new char [size];
	rd->setPos (d.sh_offset);

	for (int i = 0; i < size; ++i)
	  {
	    data[i] = rd->get8 ();
	  }
      }
	
    return (true);
  }

void SectionHeader::print (ostream &strm)
  {
    strm << "  File pos " << asHex32X (filePos) << '\n';
    strm << "  Name:    " << d.sh_name << " - ";
    strm << &shtringer [d.sh_name] << '\n';
    strm << "  Type:    " << d.sh_type << '\n';
    strm << "  Flags:   " << d.sh_flags << '\n';
    strm << "  Address: " << d.sh_addr << '\n';
    strm << "  Offset:  " << asHex32X (d.sh_offset) << '\n';
    strm << "  Size:    " << d.sh_size << '\n';
    strm << "  Link:    " << d.sh_link << '\n';
    strm << "  Info:    " << d.sh_info << '\n';
    strm << "  Align:   " << d.sh_addralign << '\n';
    strm << "  EntSize: " << d.sh_entsize << '\n';

    if (dis)
      {
	dis->print (strm);
      }
    if (data)
      {
	int size = d.sh_size;
	int cnt = 0;
	strm << "Raw content:";
	while (cnt < size)
	  {
	    strm << "\n" << asHex32X (d.sh_offset + cnt) << " ";
	    int pcnt = cnt;
	    for (int i = 0 ; i < 16; ++i)
	      {
		if (cnt < size)
		  strm << asHex8 (data[cnt++]);
		else
		  strm << "--";
		strm << " ";
	      }
	    for (int j = 0 ; j < 16; ++j)
	      {
		if (pcnt < size)
		  {
		    if (data[pcnt] > ' ' && data[pcnt] < 0x7F)
		      strm << data[pcnt];
		    else
		      strm << '.';
		    ++pcnt;
		  }
		else
		  strm << "-";
	      }
	  }
      }
    strm << "\n\n";
  }


char *SectionHeader::getName (void)
  {
/*

    static char buff[256];

    sprintf (buff, "%d", d.sh_name);
    return (buff);
    */
    return (&shtringer [d.sh_name]);
  }


//////////////////////////////////////////////////////////////////////////
//									//
//////////////////////////////////////////////////////////////////////////

ElfHeader::ElfHeader (void)
  {
    
  }

bool ElfHeader::parse (Reader *rd)
  {
    // Test for elf file here
    for (int i = 0; i < sizeof (hdr.e_ident); ++i)
      {
	hdr.e_ident[i] = rd->get8 ();
      }

    // Make some tests for file validity
    
    if (hdr.e_ident[0] == ELFMAG0 &&
	hdr.e_ident[1] == ELFMAG1 &&
	hdr.e_ident[2] == ELFMAG2 &&
	hdr.e_ident[3] == ELFMAG3)
      {
	// partial success
      }
    else
      {
	hdr.e_ident[0] = 'F';
	hdr.e_ident[1] = 'A';
	hdr.e_ident[2] = 'I';
	hdr.e_ident[3] = 'L';
	return (false);
      }
/*
    if (hdr.e_ident[EI_CLASS] != ELFCLASS32)
      {
	cout << "Supports only 32bit ELF\n";
	return (false);
      }
*/
    
    // Set reader state (lo or hi-endian) to read reminder of file
    switch (hdr.e_ident[EI_DATA])
      {
	case ELFDATA2LSB:
	  // little-endian
	  rd->bigEndian = false;
	  break;
	case ELFDATA2MSB:
	  // big-endian
	  rd->bigEndian = true;
	  break;
	case ELFDATANONE:
	default:
	  cout << "Invalid data encoding\n";
	  return (false);
      }
    if (hdr.e_ident[EI_VERSION] != EV_CURRENT)
      {
	cout << "Invalid ELF version "
	     <<(int) hdr.e_ident[EI_VERSION]
	     << '\n';
	return (false);
      }

    hdr.e_type		= rd->get16 ();
    hdr.e_machine	= rd->get16 ();
    hdr.e_version	= rd->get32 ();
    hdr.e_entry		= rd->get32 ();
    hdr.e_phoff		= rd->get32 ();
    hdr.e_shoff		= rd->get32 ();
    hdr.e_flags		= rd->get32 ();
    hdr.e_ehsize	= rd->get16 ();
    hdr.e_phentsize	= rd->get16 ();
    hdr.e_phnum		= rd->get16 ();
    hdr.e_shentsize	= rd->get16 ();
    hdr.e_shnum		= rd->get16 ();
    hdr.e_shstrndx	= rd->get16 ();
    
    return (true);
  }


bool ElfHeader::isBigEndian (void)
  {
    return (hdr.e_ident[EI_DATA] == ELFDATA2MSB);
  }

char *ElfHeader::getMachine (void)
  {
    char *ptr;
    static char buff[40];
    
    switch ((int)hdr.e_machine)
      {
	case EM_NONE:
	  ptr = "No machine";
	  break;
	case EM_M32:
	  ptr = "AT&T WE 32100";
	  break;
	case EM_SPARC:
	  ptr = "SPARC";
	  break;
	case EM_386:
	  ptr = "Intel 80386";
	  break;
	case EM_68K:
	  ptr = "Motorola 68000";
	  break;
	case EM_88K:
	  ptr = "Motorola 88000";
	  break;
	case EM_860:
	  ptr = "Intel 80860";
	  break;
	case EM_MIPS:
	  ptr = "MIPS RS3000 Big-Endian";
	  break;
	case EM_MIPS_RS4_BE:
	  ptr = "MIPS RS4000 Big-Endian";
	  break;
	case EM_PPC:
	  ptr = "PowerPC";
	  break;
	case EM_ARM:
	  ptr = "ARM";
	  break;
	default:
	  sprintf (&buff[0], "-unknown- %d", (int)hdr.e_machine);
	  ptr = buff;
      }
    return (ptr);
  }

void ElfHeader::print (ostream &strm)
  {
    strm << "ELF HEADER:\n";

    strm << "  ";
    for (int i = 1; i <= 3; ++i)
      {
	strm << hdr.e_ident[i];
      }
    strm << '\n';

    strm << "  ";
    switch (hdr.e_ident[EI_CLASS])
      {
	case ELFCLASS32:
	  strm << "32bit";
	  break;

	case ELFCLASS64:
	  strm << "64bit";
	  break;
	default:
	  strm << "Invalid class";
      }
    strm << '\n';

    strm << "  Data encoding: ";
    switch (hdr.e_ident[EI_DATA])
      {
	case ELFDATA2LSB:
	  strm << "ELFDATA2LSB (little-endian)";
	  break;
	case ELFDATA2MSB:
	  strm << "ELFDATA2MSB (big-endian)";
	  break;
	case ELFDATANONE:
	default:
	  strm << "invalid";
      }
    strm << '\n';
    
    strm << "  Version: " << "header " <<(int) hdr.e_ident[EI_VERSION]
	 << "  -  e_version " << (int)hdr.e_version <<'\n';

    strm << "  Machine: " << getMachine () <<'\n';
    strm << "  Entry: " << (int)hdr.e_entry <<'\n';
    strm << "  Program header table's file offset: " << (int)hdr.e_phoff <<'\n';
    strm << "  Section header table's file offset: " << (int)hdr.e_shoff <<'\n';
    strm << "  Flags: " <<  (int)hdr.e_flags <<'\n';
    strm << "  Elf header size: " << (int)hdr.e_ehsize <<'\n';
    strm << "  Size of program header entry: " << (int)hdr.e_phentsize <<'\n';
    strm << "  No of entries in program header table: " << (int)hdr.e_phnum <<'\n';
    strm << "  Size of section header entry: " << (int)hdr.e_shentsize <<'\n';
    strm << "  No of entries in section header table: " << (int)hdr.e_shnum <<'\n';
    strm << "  Section header table index of string table: " << (int)hdr.e_shstrndx <<'\n';
    strm << '\n';
  }


//////////////////////////////////////////////////////////////////////////
//									//
//////////////////////////////////////////////////////////////////////////

ElfFile::ElfFile (Reader *in)
  {
    unsigned long l;

    strIndex =0;
    secHdrs = 0;
    rd = in;

    elfHdr = new ElfHeader ();

    if (elfHdr->parse (rd) == false)
      {
	isValid = false;
	return;
      }
    isValid = true;
    bigEndian = elfHdr->isBigEndian ();

    secHdrs = new (SectionHeader (* [elfHdr->hdr.e_shnum]));
    int filestart = elfHdr->hdr.e_shoff;
    int shentsize = elfHdr->hdr.e_shentsize;
    int filepos = filestart;

    numHdrs = elfHdr->hdr.e_shnum;
    
    for (int i = 0 ; i < numHdrs; ++i)
      {
	rd->setPos (filepos);
	secHdrs[i] = new SectionHeader ();
	secHdrs[i]->parse (rd);
	filepos += shentsize;
      }
    strIndex = secHdrs[elfHdr->hdr.e_shstrndx]->d.sh_offset;
    shtringer = secHdrs[elfHdr->hdr.e_shstrndx]->data;

    for (int j = 0 ; j < numHdrs; ++j)
      {
	if (j == elfHdr->hdr.e_shstrndx)
	  {
	    sh_sect = new SHSection (secHdrs[j]->data,
				     secHdrs[j]->d.sh_size);
	    secHdrs[j]->dis = sh_sect;
	  }
	if (strcmp (".strtab",&shtringer [secHdrs[j]->d.sh_name]) == 0)
	  {
	    stringer = secHdrs[j]->data;
	  }
	if (strcmp (".symtab", &shtringer [secHdrs[j]->d.sh_name]) == 0)
	  {
	    secHdrs[j]->dis = sym_tab = new SymbolTable (secHdrs[j]->data,
					       secHdrs[j]->d.sh_size,
					       secHdrs[j]->d.sh_size/secHdrs[j]->d.sh_entsize );
	  }
	if (strcmp (".rel.text", &shtringer [secHdrs[j]->d.sh_name]) == 0)
	  {
	    secHdrs[j]->dis = new RelSection (secHdrs[j]->data,
					      secHdrs[j]->d.sh_size,
					      secHdrs[j]->d.sh_entsize );
	  }
	if (strcmp (".text", &shtringer [secHdrs[j]->d.sh_name]) == 0)
	  {
	    secHdrs[j]->dis = new TextSection (secHdrs[j]->data,
					       secHdrs[j]->d.sh_size);
	  }
      }
    
  }



void ElfFile::print (ostream &strm)
  {
    if (isValid)
      {
	cout << "String index: " << asHex32X (strIndex) << '\n';
		

	strm << "ELF FILE\n";
	elfHdr->print (strm);
	if (secHdrs)
	  {
	    for (int i = 0; i < numHdrs; ++i)
	      {
		strm << "SECTION HEADER: " << i << "\n";

		secHdrs[i]->print (strm);
	      }
	  }
      }
    else
      {
	strm << "Not a valid ELF file\n";
      }
  }


//////////////////////////////////////////////////////////////////////////
//									//
//////////////////////////////////////////////////////////////////////////

Helper::Helper (void)
  {
    bigEndian = false;
  }

void Helper::setBigEndian (void)
  {
    bigEndian = true;
  }

int Helper::makeHalfWord (unsigned char b1,
			  unsigned char b2)
  {
    int tmp;
    
    if (bigEndian)
      {
	tmp = b1 * 256 + b2;
      }
    else
      {
	tmp = b2 * 256 + b1;
      }
    return (tmp);
  }

int Helper::makeWord (unsigned char b1,
		      unsigned char b2,
		      unsigned char b3,
		      unsigned char b4)
  {
    int tmp;
    
    if (bigEndian)
      {
	tmp = makeHalfWord (b1, b2) * 256 * 256 + makeHalfWord (b3, b4);
      }
    else
      {
	tmp = makeHalfWord (b1, b2) + makeHalfWord (b3, b4) * 256 * 256;
      }
    return (tmp);
  }


//////////////////////////////////////////////////////////////////////////
//									//
//////////////////////////////////////////////////////////////////////////

void error (const char *fmt, ...)
  {
    va_list args;

    va_start (args, fmt);

    vprintf (fmt, args);
    va_end (args);
    printf ("\n");
  }


void usage (void)
  {
    printf ("usage:\n");
    printf ("   disarm filename\n");
  }

//////////////////////////////////////////////////////////////////////////
//									//
//////////////////////////////////////////////////////////////////////////

int main (int argc, char *argv[])
  {
    int loc_add;

    if (argc < 2)
      {
	error ("No file given");
	usage ();
	return (2);
      }
    
    FileBuff *f = new FileBuff (argv[1]);
    Reader *r = new Reader (f);
    
//    AOFFile *aof = new AOFFile (r);
    ElfFile *eof = new ElfFile (r);
    g_elf = eof;
    
//    g_aof = aof;

    cout << "Object file dump of '" << argv[1] << "'\n";

    eof->print (cout);

    cout << "END\n";
    
    return (0);
  }
